##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

# Using a single configuration file with all paths and relevant options
# The rest of the Tasmanian Python modules will not be configured by cmake

__version__ = "@Tasmanian_VERSION_MAJOR@.@Tasmanian_VERSION_MINOR@"
__license__ = "@Tasmanian_license@"
__author__ = "Miroslav Stoyanov"
__git_commit_hash__ = "@Tasmanian_git_hash@"

__path_libsparsegrid__ = "@Tasmanian_libsparsegrid_path@"
__path_libdream__      = "@Tasmanian_libdream_path@"
__path_libcaddons__    = "@Tasmanian_libcaddons_path@"
__path_logfile__       = "@Tasmanian_final_install_path@/share/Tasmanian/Tasmanian.log"

enableVerboseErrors = True

# the transitive library dependencies on Linux/OsX are handled by rpath, Windows needs help
if "@CMAKE_SYSTEM_NAME@" == "Windows":
    import os
    os.environ['PATH'] = "@Tasmanian_libpaths@" + os.pathsep + os.environ['PATH']

class TasmanianInputError(Exception):
    '''Exception raised for incorrect input to Tasmanian

    Attributes:
    sVariable   -- string containing the variable name with incorrect value
    sMessage    -- message regarding the error
    bShowOnExit -- boolean indicates whether to call printInfo() form the destructor

    '''
    def __init__(self, sVar, sMess):
        self.sVariable = sVar
        self.sMessage = sMess
        self.bShowOnExit = enableVerboseErrors

    def __del__(self):
        if self.bShowOnExit:
            self.printInfo()

    def printInfo(self):
        '''
        prints information for the incorect function or variable

        '''
        print("Incorrect input for: {0:s}".format(self.sVariable))
        print(self.sMessage)
